package org.flixel.plugin.FlxSWFStates
{
	
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.BlendMode;
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.display.Sprite;
	import flash.display.Stage;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.geom.Matrix;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.system.ApplicationDomain;
	import flash.utils.getQualifiedClassName;
	import flash.utils.getQualifiedSuperclassName;
	
	import org.flixel.*;
	
	public class SWFStatePopulator extends EventDispatcher
	{
		protected var _stateToPopulate:FlxState;
		private var _onPopulatedCallback:Function = null;
		
		public function SWFStatePopulator(stateToPopulate:FlxState,onPopulatedCallback:Function = null)
		{
			_stateToPopulate = stateToPopulate;
			
			_onPopulatedCallback = onPopulatedCallback;
			
		}
		
		protected function processStageInstance( mapStage:Sprite ):void
		{
			for( var i:int=0; i < mapStage.numChildren; i++ )
			{
				var obj:DisplayObject = mapStage.getChildAt( i );
				
				processStageObject( obj, mapStage );
				
			}
			
			var worldBounds:Rectangle = mapStage.getBounds( mapStage );
			FlxG.worldBounds = new FlxRect( worldBounds.x, worldBounds.y, worldBounds.width, worldBounds.height );
			
			dispatchEvent( new Event( Event.COMPLETE ) );
			
			if( _onPopulatedCallback != null )
			{
				// Call the callback function.
				_onPopulatedCallback.call( null, mapStage );
			}
			
		}		
		
		private function processStageObject( obj:DisplayObject, mapStage:Sprite ):void
		{
			if( obj is ObjectProxyParent )
			{
				var doc:DisplayObjectContainer = obj as DisplayObjectContainer;
				for( var i:int=0; i < doc.numChildren; i++ )
				{
					// RECURSION :O
					processStageObject( doc.getChildAt( i ), mapStage );
				}
				return;
			}
			
			
			var newInstanceClass:Class = null;
			
			if( obj.hasOwnProperty( "className" ) && obj["className"] != null && obj["className"] != "" )
			{
				if( ApplicationDomain.currentDomain.hasDefinition( obj["className"] ) )
				{
					// If the proxy was a component with "className" defined, use that class name for the object.
					newInstanceClass = ApplicationDomain.currentDomain.getDefinition( obj["className"] ) as Class;
				}
				else
				{
					// The proxy attempted to define a class name but no valid one was found! Throw an error indicating this
					FlxG.log("WARNING: Stage instance " + obj + " with name " + obj.name + " defined className " + obj["className"] + " but no such class was found!" );
				}
			}
			
			var instanceNameAsClassName:String = obj.name;
			
			if( instanceNameAsClassName != null && newInstanceClass == null )
			{
				// Trim trailing $'s that exist sometimes for some reason.
				instanceNameAsClassName = instanceNameAsClassName.replace( /\$/g,".");
				
				// These ID's are auto-generated and meaningless.
				var re:RegExp = new RegExp( "__id(\\d+)_" );
				if( instanceNameAsClassName.match( re ) )
				{
					instanceNameAsClassName = "";
				}
				
			}
			
			if( newInstanceClass == null && instanceNameAsClassName != null && instanceNameAsClassName != "" ) 
			{
				if( ApplicationDomain.currentDomain.hasDefinition( instanceNameAsClassName ) )
				{
					// If the proxy wasn't a component or had no class name defined, maybe someone just didn't want the trouble
					//	of making it a Component, but still put the class name they wanted to use as the instance name of the object.
					newInstanceClass = ApplicationDomain.currentDomain.getDefinition( instanceNameAsClassName ) as Class;
				}
				else
				{
					// The proxy attempted to define a class name but no valid one was found! Throw an error indicating this
					FlxG.log("WARNING: Stage instance " + obj + " has instance name " + obj.name + ", massaged into " + instanceNameAsClassName + " - but this does not match a class name!" );
				}
			}
			
			if( newInstanceClass == null )
			{
				// If there seems to have been no attempt to specify the class of this thing, then just make it a FlxSprite.
				newInstanceClass = org.flixel.FlxSprite;
			}
			
			
			
			// Now we initialize this object with the fields and values specified on the proxy.
			var newInstance:Object = new newInstanceClass();
			
			// First initialize the x, y, width, and height based on the object's position on the stage.
			//	TODO: Rotation too?
			
			var objectProxyBounds:Rectangle = obj.getBounds( mapStage );
			
			if( newInstance is FlxObject )
			{
				FlxObject( newInstance ).reset( objectProxyBounds.x, objectProxyBounds.y );
			}
			else
			{
				initPropertyOnInstance( newInstance, "x", 		objectProxyBounds.x 		);
				initPropertyOnInstance( newInstance, "y", 		objectProxyBounds.y 		);
			}
			
			initPropertyOnInstance( newInstance, "width",	objectProxyBounds.width 	);
			initPropertyOnInstance( newInstance, "height",	objectProxyBounds.height	);
			
			if( obj.blendMode != BlendMode.NORMAL )
			{
				initPropertyOnInstance( newInstance, "blend",	obj.blendMode			);
			}
			
			if( obj.hasOwnProperty( "initValues" ) && obj["initValues"] != null )
			{
				var initValues:Vector.<DataProxy_NameValPair> = obj["initValues"].items;
				for each ( var nameValPair:DataProxy_NameValPair in initValues )
				{
					initPropertyOnInstance( newInstance, nameValPair.name, nameValPair.value );
				}
			}
			
			// If the class implements a function to be called after it's initialized from a proxy, then
			//	call it.
			if( newInstance.hasOwnProperty("postProxyInit") )
			{
				newInstance.postProxyInit();
			}
			
			// Set the pixels to match what was on the SWF stage.
			if	( !obj.hasOwnProperty( "usePixels" ) || 
				( obj.hasOwnProperty( "usePixels" ) && Boolean( obj.hasOwnProperty( "usePixels" ) ) == true ) 
			)
			{
				var collisionBoxRect:Rectangle = null;
				
				// First, if there was an alternate collision bounding-box defined within the symbol, let's respect that.
				if( obj is DisplayObjectContainer )
				{
					var collisionBox:DisplayObject = ( obj as DisplayObjectContainer ).getChildByName( "collisionBox" );
					
					if( collisionBox != null )
					{
						collisionBox.visible = false;	// Ensure it doesn't show up in the drawing we'll do below.
						
						collisionBoxRect = collisionBox.getBounds( obj );
					}
				}
				
				var localObjectProxyBounds:Rectangle = obj.getBounds( obj.parent );
				
				var localToGlobalScaleX:Number = 1;
				var localToGlobalScaleY:Number = 1;
				
				var tmpParent:DisplayObject = obj.parent;
				while( tmpParent != null )
				{
					localToGlobalScaleX *= tmpParent.scaleX;
					localToGlobalScaleY *= tmpParent.scaleY;
					tmpParent = tmpParent.parent;
				}
				
				
				var bmpd:BitmapData = new BitmapData( Math.max( 1, Math.ceil( obj.width * localToGlobalScaleX ) ), Math.max( 1, Math.ceil( obj.height * localToGlobalScaleY ) ), true, 0xffffff);
				var matrix:Matrix = obj.transform.matrix;
				matrix.translate( -localObjectProxyBounds.x, -localObjectProxyBounds.y );
				matrix.scale( localToGlobalScaleX, localToGlobalScaleY );
				
				bmpd.draw( obj, matrix );
				
				// Set the pixels.
				if( newInstance.hasOwnProperty("setProxyBitmapData") )
				{
					newInstance.setProxyBitmapData( bmpd, collisionBoxRect );
				}
				else if( newInstance.hasOwnProperty("pixels") )
				{
					newInstance.pixels = bmpd;
					
					if( collisionBoxRect != null )
					{
						newInstance.offset.x = collisionBox.x;
						newInstance.offset.y = collisionBox.y;
						
						newInstance.width = collisionBox.width;
						newInstance.height = collisionBox.height;
					}
					
				}
				
			}
			
			// Finally, actually add this thing to the scene!
			if( newInstance is FlxBasic )
			{
				if( obj.hasOwnProperty( "className" ) && obj["className"] == "LevelProperties" )
				{
					// Apply special level properties.
					for each ( var pair:DataProxy_NameValPair in obj["initValues"].items as Vector.<DataProxy_NameValPair> )
					{
						switch( pair.name )
						{
							case( "bgcolor" ):
								FlxG.bgColor = hexStringToUint( pair.value );
								break;
						}
					}
					
				}
				else if( _stateToPopulate.hasOwnProperty("addObjectFromSWFProxy") )
				{
					// If the class has exposed a replacement to the regular add() function for us to use instead, then... use it instead.
					_stateToPopulate["addObjectFromSWFProxy"]( newInstance, obj );
				}
				else
				{
					_stateToPopulate.add( newInstance as FlxBasic );
				}
			}
		}
		
		protected function initPropertyOnInstance(newInstance:Object, name:String, value:*):void
		{
			var firstDotIndex:int = name.indexOf(".");
			if( firstDotIndex != -1 )
			{
				// This is referring to a subproperty.
				
				var firstElement:String = name.substr(0,firstDotIndex);
				var rest:String = name.substr(firstDotIndex+1);
				
				if( newInstance.hasOwnProperty( firstElement ) )
				{
					initPropertyOnInstance( newInstance[ firstElement ], rest, value );
				}
				
			}
			else if( newInstance.hasOwnProperty( name ) )
			{
				newInstance[ name ] = value;
			}
		}
		
		static public function hexStringToUint( hex:String ):uint
		{
			if( hex.indexOf( "#" ) == 0 )
			{
				hex = hex.slice( 1 );
			}
			
			if( hex.indexOf( "0x" ) == 0 )
			{
				hex = hex.slice( 2 );
			}
			
			return parseInt( hex, 16 );
		}
		
	}
}